c*********************************************************************
      program gauss
c**********************************************************************
c
c     dummy program using fast gauss transform
c     main documentation is in subroutine gausst (below).
c
c     compile with the following lines:

cc    f77 -c gausst.f
cc    f77 gdriver.f gausst.o second.o 

c     input:
c
c     xat(i) = x-component of center of ith gaussian.
c     yat(i) = y-component of center of ith gaussian.
c     str(i) = height of ith gaussian.
c     natoms = number of gaussians.
c     delta  = parameter determining gaussian width.
c
c     output:
c                m
c     pot(j) =  sum  str(i) * exp[((xat(i) - xat(j))^2 +
c               i=1                (yat(i) - yat(j))^2)/delta ]
c
c----------------------------------------------------------------------
c     include statement for graphics to draw points in box
c      include '/usr/include/f77/cgidefs77.h'
c
c-----variables
c
      implicit real *8  (a-h,o-z)
      parameter(maxwrk=1000000)
      parameter(maxatm=110000)
      integer*4  natoms,lenw,i,iflag,iflag2,imax,ier(20),inform(20)
      integer*4  iout1,iout2,ictot,iwork(maxwrk)
      real *8  xat(maxatm),yat(maxatm),pot(maxatm),str(maxatm)
      real *8  xtarg(maxatm),ytarg(maxatm)
      real *8  work(maxwrk),tol,ctot,eps,dirpot(200)
c
c-----set parameters
c
      iout1 = 6
      iout2 = 13
      lenw = maxwrk
ccc   call prini(iout1,iout2)
100   continue
      print *,'enter empty line and then icase '
      read *
      read *, icase
      print *,' icase = ',  icase
      write(13,*) ' icase = ', icase
      if(icase.eq.0)goto 1111

      delta = .01
      print *,' enter delta '
      read *,delta
      print *, ' delta = ',delta
      write(13,*) ' delta = ',delta
      natoms = 20
      ntarg = 20
      print *,' enter natoms '
      read *,natoms
      print *, ' natoms = ',natoms
      write(13,*) ' natoms = ',natoms

      print *,' enter ntarg '
      read *,ntarg
      print *, ' ntarg = ',ntarg
      write(13,*) ' ntarg = ',ntarg

      tol = 1.d-9
      print *,' enter tol '
      read *,tol
      print *, ' tol = ',tol
      write(13,*) ' tol = ',tol
c
c-----obtain initial coordinates and strengths
c

      iseed=34343
      call dummy(icase,iseed,natoms,xat,yat,str)

c     evaluate total charge in l1 norm
      q=0.0d0
      do 345 j=1,natoms
      q=q+abs(str(j))
345   continue

c     compute radius beyond which gaussian interactions
c     are less than tol
      radius=dsqrt(-delta*dlog(tol))
      write(6,*) ' interaction radius is ',radius
      write(13,*) ' interaction radius is ',radius
cc      iii=1
cc      write(6,*) ' enter 1 to have circles plotted, 0 for points'
cc      read(5,*) iii
cc      radius=radius*iii

c     plot sources as circles with this radius
ccc   call drawpts(1,natoms,xat,yat,radius)

      iseed=65573
      call dummy(icase,iseed,ntarg,xtarg,ytarg,str)
c     write(6,*)(xtarg(i),i=1,ntarg)
c     write(6,*)(ytarg(i),i=1,ntarg)

c     plot targets as points
      radius=0.0d0
ccc   call drawpts(2,ntarg,xtarg,ytarg,radius)

c     compute number of terms necessary to achieve absolute error tol

      
      toler=tol
      boxdim = dsqrt(0.5d0*delta)
      nside = int(1.0d0/boxdim)+1
      boxside=1.d0/(float(nside)*sqrt(2.0*delta))
      ip=1
      z= 1.09**2*boxside**4/(1.-boxside)**2
645   continue
      ip=ip+1
      z=boxside*boxside*z/float(ip)
      if(z.gt.toler)goto 645
cc      write(6,*) 'z= ',z,' toler = ',toler,' q= ',q,' ip = ',ip
      nterms=ip
      print *,' we recommend nterms =  ',nterms
      write(13,*) ' we recommend nterms =  ',nterms
      print *,' enter nterms'
      read *, nter
      if(nter.ne.0)nterms=nter
      print *, ' nterms = ',nterms
      write(13,*) ' nterms = ',nterms


c
c-----compute field for all particles.
c

      t1 = second()
      call gausst(xat,yat,str,natoms,
     >            xtarg,ytarg,pot,ntarg,
     >            delta,tol,nterms,lenw,
     >            iwork,work,iout1,iout2,inform,ier)
      t2 = second()
c
c----- compare with direct calculation
c
      imax = min(ntarg,20)
      write(6,*)' fast potentials '
      write(13,*)' fast potentials '
      write(6,*)(pot(i),i=1,imax)
      write(13,*)(pot(i),i=1,imax)
c
c-----compute fields at first imax particle positions directly.
c
      salg = 0.0d0
      sdir = 0.0d0
      stot = 0.0d0
      smax=0.0d0
      t3 = second()
      do 700 i = 1,imax
	 xp = xtarg(i)
	 yp = ytarg(i)
	 call gslow(xp,yp,xat,yat,natoms,str,dirpot(i),delta)
	 salg = salg + (pot(i) - dirpot(i))**2
	 stot = stot + dirpot(i)*dirpot(i)
	 smax=max(abs(pot(i)-dirpot(i)),smax)
 700  continue
      t4 = second()
      write(6,*)' direct potentials '
      write(13,*)' direct potentials '
      write(6,*)(dirpot(i),i=1,imax)
      write(13,*)(dirpot(i),i=1,imax)

      write(6,*)' time for ',natoms,ntarg,' fast is ',t2-t1
      write(13,*)' time for ',natoms,ntarg,' fast is ',t2-t1
      write(6,*)' time for ',natoms,imax,' slow is ',t4-t3
      write(13,*)' time for ',natoms,imax,' slow is ',t4-t3
      tslow=(ntarg/imax)*(t4-t3)
      write(6,*)' time for ',natoms,ntarg,' slow would be ',
     >         (ntarg/imax)*(t4-t3)
      write(13,*)' time for ',natoms,ntarg,' slow would be ',
     >         (ntarg/imax)*(t4-t3)

      ealg=dsqrt(salg/stot)
      write(6,*)' l2 error is ',ealg
      write(13,*)' l2 error is ',ealg
      write(6,*)' absolute error is ',smax
      write(13,*)' absolute error is ',smax
      write(6,*)' absolute error/q is ',smax/q
      write(13,*)' absolute error/q is ',smax/q
      write(15,*)
      write(15,*) ' delta             p     tol    '
      write(15,1001) delta,nterms,tol
      write(15,*) '   n     m     fast    direct   error/q    r     i '

c     make table entry
      write(17,1009) icase,natoms,ntarg,t2-t1,tslow,smax/q,
     >           inform(4),inform(3)
1009  format(1x,i2,' & ',i6,' & ',i6,' & ',f8.3,' & ',f10.3,' & ',
     >           ' & ',e8.3,' & ',i8,' & ',i8,' \\\\ ')

1001  format(2x,f10.6,i4,e12.4)
1003  format(2x,i8,2x,i8,f12.3,2x,f15.3,2x,e12.3,2x,i8,2x,i8)
c
ccc   call drawstop
      goto 100
1111  continue
      stop
      end
c
c
c**********************************************************************
      subroutine dummy(icase,iseed,natoms,xat,yat,str)
c
c     sets random initial conditions using unix random number
c     generator..... machine independent ...
c
c**********************************************************************
      implicit real *8 (a-h,o-q,s-z)
      integer*4 natoms
      real *8 xat(natoms),yat(natoms),str(natoms)
c
c-----local variables
c
      integer*4 j
c
      if(icase.eq.1)then
      print *,' uniform distribution in a box '
      write(13,*) ' uniform distribution in a box '
c-----distribute molecules randomly in box and assign random charges.
c
      idum=-1
      x=ran1(idum)
      do 1000 j = 1,natoms
	 x=ran1(idum)
	 xat(j) = x
	 y=ran1(idum)
	 yat(j) = y
	 s=ran1(idum)
	 str(j) = (-1)**j*s
1000  continue

      else if(icase.eq.2)then

      print *,' uniform distribution on a circle '
      write(13,*) ' uniform distribution on a circle '

      radius=0.23456d0
      xcenter=0.5d0
      ycenter=0.5d0
      ds=3.1415926535d0*2.d0/float(natoms)
      do 2000 j=1,natoms
      xat(j)=xcenter+radius*dcos(j*ds)
      yat(j)=ycenter+radius*dsin(j*ds)
      str(j)=ds*dcos(j*ds)
ccc      write(6,*)  ' x,y,str ',xat(j),yat(j),str(j)
2000  continue

      end if
      return
      end
c
c*********************************************************************c
      subroutine gslow(xp,yp,xat,yat,natoms,str,dpot,delta)
c
c---- direct calculation
c
      implicit real *8  (a-h,o-z)
      real *8 xat(1),yat(1),str(1),dpot,xp,yp,delta
c
      dpot = 0.0d0
      do 30 j = 1,natoms
	 x = xp - xat(j)
	 y = yp - yat(j)
	 dpot = dpot + dexp(- (x**2 + y**2)/delta)*str(j)
30    continue
      return
      end
      real*8 function ran1(idum)
      implicit real*8(a-h,p-z)

      dimension r(97)
      parameter(m1=259200,ia1=7141,ic1=54773,rm1=3.8580247d-6)
      parameter(m2=134456,ia2=8121,ic2=28411,rm2=7.4373773d-6)
      parameter(m3=243000,ia3=4561,ic3=51349)
      data iff /0/

      if(idum.lt.0.or.iff.eq.0)then
	 iff=1
	 ix1=mod(ic1-idum,m1)
	 ix1=mod(ia1*ix1+ic1,m1)
	 ix2=mod(ix1,m2)
	 ix1=mod(ia1*ix1+ic1,m1)
	 ix3=mod(ix1,m3)
	 do 11 j=1,97
	    ix1=mod(ia1*ix1+ic1,m1)
	    ix2=mod(ia2*ix2+ic2,m2)
	    r(j)=(float(ix1)+float(ix2)*rm2)*rm1
11       continue
         idum=1
      end if
      ix1=mod(ia1*ix1+ic1,m1)
      ix2=mod(ia2*ix2+ic2,m2)
      ix3=mod(ia3*ix3+ic3,m3)
      j=1+(97*ix3)/m3
      if(j.gt.97.or.j.lt.1)write(6,*) ' error in ran1 '
      ran1=r(j)
      r(j)=(float(ix1)+float(ix2)*rm2)*rm1
      return
      end
      
c**********************************************************************c
      subroutine gausst(xat,yat,str,natoms,
     >                  xtarg,ytarg,pot,ntarg,
     >                  delta,tol,nterms,lenw,
     >                  iwork,work,iout1,iout2,inform,ier)
c**********************************************************************c
c
c     description :
c
c     this subroutine computes the fast gauss transform, that is
c     the sum of gaussians of the form
c
c                m
c     pot(j) =  sum  str(i) * exp[((xat(i) - xtarg(j))^2 +
c               i=1                (yat(i) - ytarg(j))^2)/delta ]
c
c     the source positions and targets must lie within
c     the unit box [0,1] x [0,1].
c
c     tol  is the specified precision.
c          the precision which can be achieved by a double precision
c          version of this program can not be set higher than about
c          1.0d-14. if too high a precision is requested, the
c          subroutine returns with an error code described below.
c
c     nterms = number of terms desired. if nterms = 0, an estimate
c          is made depending on tol.
c
c     work is a workspace array dimensioned as real *8.
c          the amount of workspace required is approximately
c          ( 25 + log_2(tol) )*natoms.
c          if insufficient workspace is provided, the subroutine
c          returns with an error code described below.
c
c     iwork is a workspace array dimensioned as integer*4.
c
c     iout1,iout2 are logical unit numbers for printing output.
c          all reports generated during the execution of the subroutine
c          are printed on two fortran units iout1, iout2 specified
c          by the user. 
c
c     on  input :
c
c     xat(i) = center of ith gaussian
c     yat(i) = center of ith gaussian
c     str(i) = hieght of ith gaussian.
c     natoms = number of gaussians.
c     xtarg =  x-coord of ith target
c     ytarg = y-coord of ith target
c     ntarg = number of targets
c     work   = real work array
c     iwork  = integer work array
c     lenw       = length of work arrays
c     tol        = desired precision of calculation
c     nterms = number of terms desired
c
c     on  output :
c
c     pot(i) = potential at position of ith gaussian.
c
c     inform   = information returned to user
c
c              inform(1) is the number of levels actually used
c                 in the calculations.
c              inform(2) is the total number of boxes used in the
c                 calculations.
c              inform(3) is the amount of integer workspace actually used
c                 in the calculations.
c              inform(4) = amount of real workspace
c              inform(5) is the number of terms used in the
c                 hermite expansions to achieve desired accuracy.
c              inform(7,8,9,10) are the numbers of interactions handled 
c              by direct, accum-taylor, hermite-eval, and hermite-taylor
c              methods respectively
c
c     ier        = error code array
c
c     ier(1) = 0  => no errors encountered
c
c     ier(1) = 4  => tol set too high. too many terms in hermite
c                    expansions needed.
c                    ier(2) returns the number of terms in expansion
c                    required to satisfy tol, which should not exceed
c                    20.
c
c     ier(1) = 8  => insufficient workspace allotted.
c                    ier(2) returns amount of workspace needed
c                    ier(3) returns amount of workspace provided.
c
c     (local variables)
c
c     nterms     = number of terms used in expansions
c
c**********************************************************************c
      implicit real *8  (a-h,o-z)
      integer*4 natoms,lenw,iout1,iout2,ier(1)
      integer*4 iflag,iflag2,inform(1),iwork(1)
      real *8  xat(1),yat(1),str(1),pot(1),tol,work(1),delta
      real *8  xtarg(1),ytarg(1)

      tol=max(1.d-12,tol)
      tol=min(1.d-1,tol)
      boxdim = dsqrt(0.5d0*delta)
      nmax = max(natoms,ntarg)
      nside = int(1.0d0/boxdim) +1

      ier(1) = 0
      if (nterms .eq. 0) then
c
c----- determine number of hermite expansion terms needed for
c      specified precision
c
      toler=tol
      boxside=1.d0/(float(nside)*sqrt(2.0*delta))
      ip=0
      z= 1.09**2*boxside**4/(1.-boxside)**2
645   continue
      ip=ip+1
      z=boxside*boxside*z/float(ip)
      if(z.gt.toler)goto 645
      nterms=ip

      end if
      inform(5) = nterms
      write(iout1,*)' #terms in hermite expansions  =',nterms
      write(iout2,*)' #terms in hermite expansions  =',nterms
      if (nterms .gt. 20) then
	 ier(1) = 4
      endif
c
c----- determine total number of boxes and set boxsize to 64:
c

      kdis = dsqrt(-2.d0*dlog(tol))+1
      nfmax = nterms+2
      nlmax = nterms+2
      write(iout1,*)' krange = ',kdis
      write(iout2,*)' krange = ',kdis
      nallbx = nside*nside
      inform(2) = nallbx
      write(iout1,*)' # of boxes = ',nallbx
      write(iout2,*)' # of boxes = ',nallbx
c
c----- allocate arrays used by algorithm from workspace.
c
      lfarxp = (nterms+1)*(nterms+1)
      licnt  = nallbx
      licnt2 = nallbx
      lloc   = nallbx
      ltofst = nallbx+1
      loffst = nallbx+1
      lcentr = 2*nallbx
      libox  = nmax
      ltradr = ntarg
      lxatms = natoms
      lyatms = natoms
      lchrg2 = natoms
      lbexp = (nterms+1)*(nterms+1)
      ltemp = (nterms+1)*(nterms+1)
      lnbors = nallbx
c
      icnt = 1
      icnt2 = icnt + licnt
      ioffst = icnt2 + licnt2
      itofst = ioffst + loffst
      ibox = itofst + ltofst
      itradr = ibox + libox
      iloc = itradr + ltradr
      inbors = iloc + lloc
      itot = iloc + lnbors
c
      call lenchk(nside,xtarg,ytarg,ntarg,iwork(iloc),ntbox)
c
      llocxp = (nterms+1)*(nterms+1)*ntbox
      ifarxp =  1
      ilocxp = ifarxp + lfarxp
      ixatms = ilocxp + llocxp
      iyatms = ixatms + lxatms
      ichrg2 = iyatms + lyatms
      icentr = ichrg2 + lchrg2
      ibexp = icentr + lcentr
      itemp = ibexp + lbexp
      itotr = ibexp + ltemp
c
      if ( itotr .ge. lenw ) then
	  ier(1) = 7
	  ier(2) = itotr
	  ier(3) = lenw
      endif
      if ( itot .ge. lenw ) then
	  ier(1) = 8
	  ier(2) = itot
	  ier(3) = lenw
      endif
      inform(4) = itotr
      inform(3) = itot
      write(iout1,*)' real storage required = ',inform(4)
      write(iout2,*)' real storage required = ',inform(4)
      write(iout1,*)' integer storage required = ',inform(3)
      write(iout2,*)' integer storage required = ',inform(3)
c
c-----if error encountered, return
c
      if (ier(1) .ne. 0) then
	 call errout(ier,iout1,iout2)
	 return
      endif

      do 100 j = 1, ntarg
	 pot(j) = 0
100   continue
c
c----- call gafexp to create all expansions on grid,
c      evaluate all appropriate far field expansions,
c      and evaluate all appropriate direct interactions.
c
      call gafexp(xat,yat,str,natoms,xtarg,ytarg,ntarg,delta,nside,
     1   work(ixatms),work(iyatms),work(ichrg2),work(ifarxp),
     2   work(ilocxp),work(ibexp),work(itemp),work(icentr),nterms,
     3   tol,iwork(ioffst),iwork(ibox),iwork(itradr),
     4   iwork(itofst),iwork(icnt),iwork(icnt2),
     5   iwork(inbors),iwork(iloc),ntbox,nfmax,nlmax,pot,kdis,
     >       inform)
c
c-----evaluate all appropriate local expansions
c     (that is, for targets lying in boxes with more than nlmax
c     target points).
c
      call gaeval(xtarg,ytarg,pot,ntarg,work(ilocxp),nterms,
     1   iwork(ibox),work(icentr),delta,iwork(iloc),nlmax,nside,
     2   iwork(itofst),iwork(itradr))

      inter=inform(7)+inform(8)+inform(9)+inform(10)
      inform(7)=100*inform(7)/inter
      inform(8)=100*inform(8)/inter
      inform(9)=100*inform(9)/inter
      inform(10)=100*inform(10)/inter
      write(iout1,*) inform(7), ' % direct interactions '
      write(iout1,*) inform(8), ' % accumulated taylor series '
      write(iout1,*) inform(9), ' % hermite series evaluated '
      write(iout1,*) inform(10), ' % hermite-taylor conversions '
      write(iout2,*) inform(7), ' % direct interactions '
      write(iout2,*) inform(8), ' % accumulated taylor series '
      write(iout2,*) inform(9), ' % hermite series evaluated '
      write(iout2,*) inform(10), ' % hermite-taylor conversions '

      return
      end

c *********************************************************************c
      subroutine errout(ier,iout1,iout2)
c
c     error handling messages.
c
c
c *********************************************************************c
      integer*4 ier(1)
c
      if (ier(1) .eq. 4) then
	write(iout1,*)' error in gausst'
	write(iout1,*) ' number of terms in hermite expansions too big '
	write(iout2,*)' error in gausst'
	write(iout2,*) ' number of terms in hermite expansions too big '
      else if (ier(1) .eq. 7) then
	write(iout1,*)' error in gausst'
	write(iout1,*) ' too much real storage requested '
	write(iout2,*)' error in gausst'
	write(iout2,*) ' too much real storage requested '
      else if (ier(1) .eq. 8) then
	write(iout1,*)' error in gausst'
	write(iout1,*) ' too much integer storage requested '
	write(iout2,*)' error in gausst'
	write(iout2,*) ' too much integer storage requested '
      endif
      return
      end
c****************************************************************
      subroutine gafexp(xat,yat,str,natoms,xtarg,ytarg,ntarg,
     1       delta,nside,xatoms,yatoms,charg2,ffexp,locexp,b,
     2       temp,center,nterms,tol,ioffst,ibox,itradr,itofst,
     3       icnt,icnt2,nbors,iloc,ntbox,nfmax,nlmax,pot,kdis,
     >       inform)
c
c     the main subroutine of fast gauss algorithm.
c

c     step 1)  sources are assigned to boxes.
c
c     boxes containing sources are then dealt with sequentially.
c
c     step 2)  far field expansions are formed for each box
c              containing more than nfmax sources.
c
c     the neighboring boxes (a square of side 2*kdis)
c     are then searched for targets and processed by means of
c     following decision analysis:
c
c     if  (far field expansion has been created, i.e. there are
c          more than nfmax sources in source box) then
c
c         if (ntarg > nlmax) .. convert far field expansion to
c                               local expansion. increment inform(7)
c
c         if (ntarg .le. nlmax) .. evaluate far field expansion
c                       directly at target positions. increment inform(8)
c     else (there are fewer than nfmax sources in source box
c           and no far field expansion was created)
c
c         if (ntarg > nlmax) .. convert each source to a
c                               local expansion. increment inform(9)
c
c         if (ntarg .le. nlmax) .. evaluate gaussian field due to
c                                  sources directly. increment inform(10)
c     endif
c
c*****************************************************************
      implicit real *8 (a-h,o-z)
      integer*4 nterms,natoms,ioffst(1),ntbox
      integer*4 ibox(1),icnt(1),icnt2(1),nside,ntarg,iloc(1)
      integer*4 nfmax,nbors(1),itofst(1),itradr(1),inform(1)
c
      real *8  xat(1),yat(1),str(1),delta
      real *8  xtarg(1),ytarg(1),pot(1)
      real *8  xatoms(1),yatoms(1),charg2(1),center(2,1),cent(2)
      real *8  ffexp(0:nterms,0:nterms),locexp(0:nterms,0:nterms,1)
      real *8  b(0:nterms,0:nterms),temp(0:nterms,0:nterms)
      data zero/0.0d0/

c-----step 1) assign particles to boxes
c
      nboxes = nside*nside
      call assign(nside,xat,yat,str,xatoms,yatoms,charg2,ioffst,
     1            xtarg,ytarg,ntarg,ibox,natoms,icnt,icnt2,
     2            center,itradr,itofst)
c
c-----initialize local expansions to zero.
c
      do 70 i = 1,ntbox
	 do 60 j = 0,nterms
	    do 50 k = 0,nterms
		locexp(j,k,i) = zero
 50         continue
 60      continue
 70   continue
c
c---- process all boxes
c
      inform(7)=0
      inform(8)=0
      inform(9)=0
      inform(10)=0
c
      do 800 i = 1,nboxes
	 ioff =   ioffst(i)
	 ninbox = ioffst(i+1) - ioff
	 if ( ninbox .le. 0 ) then
	    goto 800
	 else if ( ninbox .le. nfmax ) then
c
c---- do not create far field expansion
c
	    call mknbor(i,nbors,nnbors,nside,kdis)
	    do 650 j = 1,nnbors
	       iadr = nbors(j)
	       inoff = itofst(iadr)
	       ninnbr = itofst(iadr+1) - inoff
	       if (ninnbr .le. nlmax) then
c
c---- direct interaction
c
		  do 620 k = 1,ninnbr
		     jt = itradr(inoff+k)
		     xp = xtarg(jt)
		     yp = ytarg(jt)
		     call gdir(xp,yp,xatoms(ioff+1),yatoms(ioff+1),
     1                    ninbox,charg2(ioff+1),pot(jt),delta)
620               continue
                  inform(7)=inform(7)+1
	       else
c
c---- convert each source into taylor series and accumulate
c
		  cent(1) = center(1,iadr)
		  cent(2) = center(2,iadr)
		  iadloc = iloc(iadr)
		  do 640 k = 1,ninbox
		     call galkxp(cent,xatoms(ioff+k),yatoms(ioff+k),
     1                    charg2(ioff+k),b,nterms,delta)
		     call addexp(b,locexp(0,0,iadloc),nterms)
640                continue
                  inform(8)=inform(8)+1
	       endif
650         continue
	 else
c
c---- create far field expansion
c
	    cent(1) = center(1,i)
	    cent(2) = center(2,i)
	    call gamkxp(cent,xatoms(ioff+1),yatoms(ioff+1),
     1              charg2(ioff+1),ninbox,ffexp,nterms,delta)
	    call mknbor(i,nbors,nnbors,nside,kdis)
	    do 750 j = 1,nnbors
	       iadr = nbors(j)
	       inoff = itofst(iadr)
	       ninnbr = itofst(iadr+1) - inoff
	       if (ninnbr .le. nlmax) then
c
c---- evaluate far field expansion
c
		  do 720 k = 1,ninnbr
		     jt = itradr(inoff+k)
		     xp = xtarg(jt)
		     yp = ytarg(jt)
		     call gamval(ffexp,nterms,cent,xp,yp,pot(jt),delta)
720               continue
                  inform(9)=inform(9)+1
	       else
c
c     convert far field expansion to local expansion in neighbor box
c
		  call gshift(ffexp,cent,center(1,iadr),
     1                  b,nterms,temp,delta)
		  iadloc = iloc(iadr)
		  call addexp(b,locexp(0,0,iadloc),nterms)
                  inform(10)=inform(10)+1
	       endif
750         continue
	 endif
800   continue
      return
      end
c
c**********************************************************************
      subroutine gdir(xp,yp,xat,yat,natoms,str,dpot,delta)
c**********************************************************************
c
c---- increment dpot by direct evaluation of effect of sources
c     (xat(i),yat(i),str(i))    i = 1,...,natoms.
c
      implicit real *8  (a-h,o-z)
      real *8 xat(1),yat(1),str(1),dpot,xp,yp,delta
c
      do 30 j = 1,natoms
	 x = xp - xat(j)
	 y = yp - yat(j)
	 dpot = dpot + dexp(- (x**2 + y**2)/delta)*str(j)
30    continue
      return
      end



c**********************************************************************
      subroutine addexp(b,a,nterms)
c --------------------------------------------------------------------
c     input :     taylor expansions  a and b of length nterms
c     output:     a is over written by (a+b).
c***********************************************************************
      implicit real *8 (a-h,o-z)
      integer*4  nterms
      real *8 b(0:nterms,0:nterms),a(0:nterms,0:nterms)
      integer*4 i
c
      do 200 i = 0,nterms
	 do 100 j = 0,nterms
	 a(i,j) = a(i,j) + b(i,j)
 100     continue
 200  continue
      return
      end
c
c*********************************************************************c
      subroutine assign(nside,xat,yat,str,xatoms,yatoms,charg2,ioffst,
     1              xtarg,ytarg,ntarg,ibox,natoms,icnt,icnt2,
     2              center,itradr,itofst)
c*********************************************************************c
c     this subroutine assigns sources to boxes.
c
c   input    xat(i),yat(i) is position of ith source.
c            str(i) is srtength of ith source.
c
c            icnt,icnt2 are used as workspace arrays.
c
c   output   ioffst(i) indicates where in the arrays xatoms,yatoms,
c            charg2  the listing
c            of particles contained in box i begins.
c
c            itradr is an ordered list of addresses of atoms.
c            the first group of addresses correspond to atoms which
c            lie in the first box, etc.
c
c            itofst(i) indicates where in the array itradr the listing
c            of particles contained in box i begins.
c
c            ibox(i) is the address of the box containing the ith target
c
c ------------------------------------------------------
      implicit real *8 (a-h,o-z)
      integer*4  natoms,nside,ioffst(*),ibox(natoms)
      integer*4  icnt(*),icnt2(*),itradr(*),itofst(*)
      real *8     str(natoms),charg2(natoms),center(2,1)
      real *8     xat(natoms),yat(natoms)
      real *8     xtarg(ntarg),ytarg(ntarg)
      real *8     xatoms(natoms),yatoms(natoms)
c--------------------------------------------------------
      real *8     h
c
c----- initialize counting arrays
c
      nboxes = nside*nside
      do 100 j = 1,nboxes
	 icnt2(j) = 1
	 icnt(j) = 0
 100  continue
      h = 1.0d0/nside
c
c-----find box in which jth source lies and increment counter array
c
      do 200 j = 1, natoms
	 ixh = xat(j)/h
	 iyh = yat(j)/h
	 ixh=min(ixh,nside-1)
	 iyh=min(iyh,nside-1)
	 ixh=max(ixh,0)
	 iyh=max(iyh,0)
	 icol = 1 + ixh
	 irow = 1 + iyh
	 iadr = (irow-1)*nside + icol
	 icnt(iadr) = icnt(iadr) + 1
	 ibox(j) = iadr
 200  continue
c
c-----compute the array ioffst.
c
      ioffst(1) = 0
      do 300 j = 2,nboxes+1
	 ioffst(j) = ioffst(j-1) + icnt(j-1)
 300  continue
c
c-----reorder sources in arrays xatoms,yatoms,charg2
c
      do 400 j = 1,natoms
	 iadr = ibox(j)
	 indx = ioffst(iadr) + icnt2(iadr)
	 xatoms(indx) = xat(j)
	 yatoms(indx) = yat(j)
	 charg2(indx) = str(j)
	 icnt2(iadr) = icnt2(iadr)+1
 400  continue
c
c     icnt will be used to count number of targets in boxes
c
c
      do 450 j = 1,nboxes
	 icnt2(j) = 1
	 icnt(j) = 0
 450  continue
      do 500 j = 1, ntarg
	 ixh = xtarg(j)/h
	 iyh = ytarg(j)/h
	 if (ixh .ge. nside) ixh = nside-1
	 if (iyh .ge. nside) iyh = nside-1
	 if (ixh .lt. 0) ixh = 0
	 if (iyh .lt. 0) iyh = 0
	 icol = 1 + ixh
	 irow = 1 + iyh
	 iadr = (irow-1)*nside + icol
	 icnt(iadr) = icnt(iadr) + 1
	 ibox(j) = iadr
 500  continue
c
c-----compute the array ioffst.
c
      itofst(1) = 0
      do 550 j = 2,nboxes+1
	 itofst(j) = itofst(j-1) + icnt(j-1)
 550  continue
c
c-----reorder addresses of sources in array itradr
c
      do 600 j = 1,ntarg
	 iadr = ibox(j)
	 indx = itofst(iadr) + icnt2(iadr)
	 itradr(indx) = j
	 icnt2(iadr) = icnt2(iadr)+1
 600  continue
c
c---- create array of centers for all boxes
c
      hh=0.50d0*h
      do 700 j = 1, nboxes
	 icol = 1 + mod(j-1,nside)
	 irow = 1 + (j-1)/nside
	 center(1,j) =  icol*h - hh
	 center(2,j) =  irow*h - hh
 700  continue
      return
      end
c
c
c*********************************************************************c
      subroutine gaeval(xtarg,ytarg,pot,ntarg,locexp,nterms,ibox,
     1                  center,delta,iloc,nlmax,nside,itofst,itradr)
c
c
c
c     evaluate local expansion for each particle.
c
c*****************************************************************
      implicit real *8 (a-h,o-z)
      integer*4  nterms,ntarg,itofst(1),itradr(1)
      integer*4  iloc(1)
c
      real *8  xtarg(1),ytarg(1),pot(1)
      real *8  center(2,1),cent(2)
      real *8  locexp(0:nterms,0:nterms,1)
c
      nboxes = nside*nside
      do 200 i = 1,nboxes
	 inoff = itofst(i)
	 ninbox = itofst(i+1) - inoff
	 if (ninbox .le. nlmax) then
	    goto 200
	 else
	    jadr = iloc(i)
	    do 100 k = 1,ninbox
	       jt = itradr(inoff+k)
	       xp = xtarg(jt)
	       yp = ytarg(jt)
	       call galval(locexp(0,0,jadr),nterms,center(1,i),
     1               xtarg(jt),ytarg(jt),pot(jt),delta)
100         continue
	 endif
200   continue
      return
      end
c*********************************************************************c
      subroutine gamkxp(cent,xatoms,yatoms,str,ninbox,ffexp,
     1            nterms,delta)
c
c     this subroutine computes the far field expansion about
c     the center cent due to the sources at locations
c     (xatoms(i),yatoms(i)) of strength str.
c
c     input
c
c     cent(1),cent(2) = center of the expansion
c     xatoms(i) - x-coordinate of jth source
c     yatoms(i) - y-coordinate of jth source
c     str(j)) - strength of ith source
c     ninbox - number of sources
c     nterms = number of terms in expansion
c
c     output:
c
c     ffexp(i,j) = (i,j)th coefficient of far field expansion
c                  due to sources
c
c
      implicit real *8 (a-h,o-z)
      integer*4 ninbox,nterms
      real *8 cent(2),xatoms(ninbox),yatoms(ninbox),str(ninbox)
      real *8 ffexp(0:nterms,0:nterms)
      real *8 x,y,xp,yp(0:100)
c
c     initialize coefficients to zero.
c
      do 200 i=0,nterms
	 do 100 j=0,nterms
	     ffexp(i,j) = 0.0d0
 100     continue
 200  continue
      dsq = 1.0d0/dsqrt(delta)
c
c     accumulate expansion due to each source.
c
      do 1000 i=1,ninbox
	 x = (xatoms(i) - cent(1))*dsq
	 y = (yatoms(i) - cent(2))*dsq
	 yp(0) = 1.0d0
	 do 250 k = 1,nterms
	    yp(k) = yp(k-1)*y/k
250      continue
	 xp = str(i)
	 do 400 j=0,nterms
	    do 300 k=0,nterms
	       ffexp(j,k) = ffexp(j,k) + xp*yp(k)
 300        continue
	 xp = xp*x/(j+1)
 400     continue
1000  continue
      return
      end
c
      subroutine gamval(ffexp,nterms,cent,xpoint,ypoint,pot,delta)
c
c     this subroutine evaluates the far field expansion ffexp about
c     cent at location (xp,yp).

c     input
c
c     cent(1),cent(2) = center of the expansion
c     ffexp(i,j) = (i,j)th coefficient of far field expansion
c     nterms = number of terms in expansion
c
c     output:
c
c     pot = evaluated potential
c
      implicit real *8 (a-h,o-z)
      integer*4 nterms
      real *8 cent(2)
      real *8 ffexp(0:nterms,0:nterms)
      real *8 hexpx(0:100),hexpy(0:100),x,y,xpoint,ypoint
c
c     tabulate hermite polynomials in hexpx and hexpy.
c
      dsq = 1.0d0/dsqrt(delta)
      x = (xpoint - cent(1))*dsq
      x2=2.0d0*x
      y = (ypoint - cent(2))*dsq
      y2=2.0d0*y
      facx = dexp(-x*x)
      facy = dexp(-y*y)
      hexpx(0) = facx
      hexpy(0) = facy
      hexpx(1) = x2*facx
      hexpy(1) = y2*facy
      do 100 i = 1,nterms-1
	 i2=2*i
	 hexpx(i+1) = x2*hexpx(i) - i2*hexpx(i-1)
	 hexpy(i+1) = y2*hexpy(i) - i2*hexpy(i-1)
100   continue
c
c---- evaluate expansion
c
      potinc = 0.0d0
      do 300 j = 0,nterms
	 hj=hexpx(j)
	 do 200 k = 0,nterms
	    potinc = potinc + hj*hexpy(k)*ffexp(j,k)
200      continue
300   continue
      pot = pot + potinc
      return
      end
c
c*******************************************************************
      subroutine galkxp(cent,xp,yp,charge,b,nterms,delta)
c
c     this subroutine converts the source at (xp,yp) of strength charge
c     into the taylor expansion b about cent.
c
c     input:
c
c     xp,yp  center of gaussian
c     charge strength of gaussian
c     cent = center of the box for taylor series expansion
c     nterms = number of terms in expansion
c     delta = variance
c
c     output:
c
c     b = taylor expansion coefficients
c
c
      implicit real *8 (a-h,o-z)
      integer*4 nterms
      real *8 cent(2),xp,yp,charge,delta
      real *8 b(0:nterms,0:nterms),x,y
      real *8 hexpx(0:100),hexpy(0:100)
      real *8 facx,facy,prod,fac(0:50)
c
c     tabulate hermite polynomials in hexpx and hexpy.
c
      dsq = 1.0d0/dsqrt(delta)
      x = (cent(1) - xp)*dsq
      x2=2.0d0*x
      y = (cent(2) - yp)*dsq
      y2=2.0d0*y
      facx = dexp(-x*x)*charge
      facy = dexp(-y*y)
      hexpx(0) = facx
      hexpy(0) = facy
      hexpx(1) = x2*facx
      hexpy(1) = y2*facy
      do 100 i = 1,nterms-1
	 i2=2*i
	 hexpx(i+1) = x2*hexpx(i) - i2*hexpx(i-1)
	 hexpy(i+1) = y2*hexpy(i) - i2*hexpy(i-1)
100   continue
ccc      fac(0) = 1.0d0
ccc      do 150 i = 1,nterms
ccc	 fac(i) = fac(i-1)/i
ccc150   continue
cccc
cccc---- compute local expansion
cccc
ccc      do 1600 l = 0,nterms
ccc	 do 1500 m = 0,nterms
ccc	    b(l,m) = hexpx(l)*hexpy(m)*fac(l)*fac(m)*(-1)**(l+m)
ccc1500     continue
ccc1600  continue

      facs=1.0d0
      do 1700 l=0,nterms
      hexpx(l)=hexpx(l)*facs
      hexpy(l)=hexpy(l)*facs
      facs=-facs/dble(l+1)
1700  continue
      do 1600 l = 0,nterms
	 hl=hexpx(l)
	 do 1500 m = 0,nterms
	    b(l,m) = hl*hexpy(m)
1500     continue
1600  continue
      return
      end
c
c**********************************************************************
      subroutine gshift(ffexp,cent1,cent2,local,nterms,temp,delta)
c
c     this subroutine converts the far field expansion ffexp about
c     the center cent1 to a local expansion local about cent2.
c
c     input
c
c     cent1 = center of the original expansion mpoel
c     ffexp = far field expansion coefficients
c     nterms = number of terms in expansion
c     cent2 = center about which local expansion is to be formed
c
c     output:
c
c     local(i,j) = (i,j)th coefficient of shifted local expansion
c                  due to sources
c
c
      implicit real *8 (a-h,o-z)
      integer*4 nterms
      real *8 cent1(2),cent2(2)
      real *8 ffexp(0:nterms,0:nterms),local(0:nterms,0:nterms)
      real *8 hexpx(0:100),hexpy(0:100),x,y,xpoint,ypoint
      real *8 facx,facy,sum,fac(0:50),temp(0:nterms,0:nterms)
c
c     tabulate hermite polynomials in hexpx and hexpy.
c
      dsq = 1.0d0/dsqrt(delta)
      x = (cent2(1) - cent1(1))*dsq
      x2=2.0d0*x
      y = (cent2(2) - cent1(2))*dsq
      y2=2.0d0*y
      facx = dexp(-x*x)
      facy = dexp(-y*y)
      hexpx(0) = facx
      hexpy(0) = facy
      hexpx(1) = x2*facx
      hexpy(1) = y2*facy
      do 100 i = 1,2*nterms-1
	 i2=2*i
	 hexpx(i+1) = x2*hexpx(i) - i2*hexpx(i-1)
	 hexpy(i+1) = y2*hexpy(i) - i2*hexpy(i-1)
100   continue
      fac(0) = 1.0d0
      do 150 i = 1,nterms
	 fac(i) = -fac(i-1)/i
150   continue
c
c---- compute local expansion
c
      do 600 j = 0,nterms
	 do 500 m = 0,nterms
	    sum = 0.0d0
	    do 300 k= nterms,0,-1
		  sum = sum + ffexp(j,k)*hexpy(m+k)
 300        continue
	    temp(j,m) = sum
 500     continue
 600  continue
      do 1600 l = 0,nterms
	 do 1500 m = 0,nterms
	    sum = 0.0d0
	    do 1400 j= nterms,0,-1
		  sum = sum + temp(j,m)*hexpx(l+j)
1400        continue
ccc	    local(l,m) = sum*fac(l)*fac(m)*(-1)**(l+m)
	    local(l,m) = sum*fac(l)*fac(m)
1500     continue
1600  continue
      return
      end
*****************************************************************
c
      subroutine galval(local,nterms,cent,xpoint,ypoint,pot,delta)
c
c     this subroutine evaluates the local expansion local about
c     cent at location (xpoint,ypoint).
c
c     input
c
c     cent(1),cent(2) = center of the expansion
c     local(i,j) = (i,j)th coefficient of local expansion
c     nterms = number of terms in expansion
c
c     output:
c
c     pot = evaluated potential
*****************************************************************
      implicit real *8 (a-h,o-z)
      integer*4 nterms
      real *8 cent(2),pot,parsum,potinc
      real *8 local(0:nterms,0:nterms),x,y
c
      dsq = 1.0d0/dsqrt(delta)
      x = (xpoint - cent(1))*dsq
      y = (ypoint - cent(2))*dsq
c
c---- evaluate expansion by horner's rule
c
      potinc = 0.0d0
      do 300 j = nterms,1,-1
	 parsum = 0.0d0
	 do 200 k = nterms,1,-1
	    parsum = (parsum + local(j,k))*y
200      continue
	 parsum = parsum + local(j,0)
c
	 potinc = (potinc + parsum)*x
300   continue
      parsum = 0.0d0
      do 400 k = nterms,1,-1
	 parsum = (parsum + local(0,k))*y
400   continue
      parsum = parsum + local(0,0)
      potinc = potinc + parsum
      pot = pot + potinc
      return
      end
*****************************************************************
      subroutine lenchk(nside,xtarg,ytarg,ntarg,iloc,ntbox)
c*********************************************************************c
c     this subroutine determines the amount of workspace to be reserved
c     for local expansions and sets up indirect addresses for
c     relevant boxes...
c
c     input    xtarg(i),ytarg(i) is position of ith target.
c              nside = number of boxes on a side.
c
c     output   ntbox is number of boxes containing targets.
c              iloc(j)  is pointer to location in memory of local
c              expansion for box j.
c ------------------------------------------------------
      implicit real *8 (a-h,o-z)
      integer*4  ntbox,nside,ntarg
      integer*4  iloc(*)
      real *8     xtarg(ntarg),ytarg(ntarg)
c--------------------------------------------------------
      real *8     h
c
c----- initialize counting arrays
c
      nboxes = nside*nside
      h = 1.0d0/nside
      do 100 i = 1,nboxes
	 iloc(i) = 0
100   continue
c
c-----find box in which jth target lies, increment counter
c     and compute iloc(j)
c
      do 500 j = 1, ntarg
	 ixh = xtarg(j)/h
	 iyh = ytarg(j)/h
	 if (ixh .ge. nside) ixh = nside-1
	 if (iyh .ge. nside) iyh = nside-1
	 if (ixh .lt. 0) ixh = 0
	 if (iyh .lt. 0) iyh = 0
	 icol = 1 + ixh
	 irow = 1 + iyh
	 iadr = (irow-1)*nside + icol
	 if (iloc(iadr) .eq. 0) then
	    ntbox = ntbox+1
	    iloc(iadr) = ntbox
	 endif
 500  continue
      return
      end
c
*****************************************************************
      subroutine mknbor(ibox,nbors,nnbors,nside,kdis)
c
c     creates list of neighbors of box ibox.
c
c     nside = number of boxes on a side (sqrt of total number)
c     kdis = max number of neighbors to consider in each direction.
c     ibox = box number
c     nbors = array of neighbor addresses
c     nnbors = number of neighbors to consider (avoid nonexistent
c              neighbors at boundary)
c
c*********************************************************
      integer*4 ibox,nbors(1),nnbors
c
      nnbors = 0
c
c---- compute actual col and row number of ibox
c
      irowb = (ibox-1)/nside + 1
      icolb = ibox - (irowb-1)*nside
c
c---- compute actual range (a subset of -kdis,kdis x -kdis,kdis)
c
      imin = max(icolb-kdis,1)
      imax = min(icolb+kdis,nside)
      jmin = max(irowb-kdis,1)
      jmax = min(irowb+kdis,nside)
c
c----- determine nbors list (nearest and second nearest shells).
c
      do 600 i= imin,imax
	 do 400 j= jmin,jmax
	    nnbors = nnbors+1
	    nbors(nnbors) = (j-1)*nside + i
400      continue
600   continue
      return
      end
c
